package com.emamaker.voxeltest.intervaltrees.data;

import java.util.ArrayList;
import java.util.Arrays;

public class IntervalTree<V> {

    TreeNode root = null;
    int size = 0;

    public void insertValue(V value, Interval i) {
        root = insertNode(root, new TreeNode(i, value, null, null));
        // if (isEmpty()) {
        // root = new TreeNode(i, value, null, null);
        // size++;
        // } else
        // insertValue(root, value, i);
    }

    public TreeNode getRoot() {
        return this.root;
    }

    // For what we're doing, when we add new values that (partially) overlap old
    // ones, we can't have the old ones laying around, so intervals need to be
    // readjusted accordingly
    // For a normal interval tree, there's no need no check for partial overrides
    // Also, no values can be shared from one node to another (no limits in common, e.g. one node has the high limit that is the same as the low limit of its right child)
    public TreeNode insertNode(TreeNode currentNode, TreeNode newNode) {
        if (!newNode.interval.isValid()) return null;
        if (currentNode == null)
            return newNode;

        // System.out.println(currentNode.interval.toString() +", "+
        // newNode.interval.toString());

        if (newNode.interval.fullyLeftOf(currentNode.interval)) {
            currentNode.setLeft(insertNode(currentNode.left, newNode));
            return currentNode;
        } else if (newNode.interval.fullyRightOf(currentNode.interval)) {
            currentNode.setRight(insertNode(currentNode.right, newNode));
            return currentNode;
        } else if (newNode.interval.contains(currentNode.interval)) {
            // check is they carry the same value
            if (newNode.value.equals(currentNode.value)) {
                currentNode.interval = newNode.interval;
            } else {
                TreeNode newLeftNode = new TreeNode(new Interval(newNode.interval.low, currentNode.interval.low - 1),
                        newNode.value, currentNode.left, null);

                newNode.setInterval(currentNode.interval.high + 1, newNode.interval.high);
                newNode.setRight(currentNode.right);

                // Important! otherwhise already existing nodes will node by replaced
                currentNode.value = newNode.value;

                currentNode.setLeft(newLeftNode);
                currentNode.setRight(newNode);
            }

            constrainIntervals(currentNode, currentNode.interval);
            return currentNode;

        } else if (newNode.interval.isContainedBy(currentNode.interval)) {

            if (newNode.value.equals(currentNode.value)) return currentNode;

            if (currentNode.interval.low < newNode.interval.low)
                newNode.setLeft(new TreeNode(new Interval(currentNode.interval.low, newNode.interval.low - 1),
                        currentNode.value, currentNode.left, null));
            else newNode.setLeft(null);

            if (newNode.interval.high < currentNode.interval.high)
                newNode.setRight(new TreeNode(new Interval(newNode.interval.high + 1, currentNode.interval.high),
                        currentNode.value, null, currentNode.right));
            else
                newNode.setRight(null);

            return newNode;

        } else if (newNode.interval.partiallyContainsLeft(currentNode.interval)) {
            currentNode.setInterval(currentNode.interval.low, newNode.interval.low - 1);

            newNode.setLeft(currentNode);
            constrainIntervals(currentNode, currentNode.interval);

            return newNode;
        } else if (newNode.interval.partiallyContainsRight(currentNode.interval)) {
            currentNode.setInterval(newNode.interval.high + 1, currentNode.interval.high);

            newNode.setRight(currentNode);
            constrainIntervals(currentNode, currentNode.interval);
            return currentNode;
        }

        System.out.println("How did we get here?");
        return null;
    }

    // Get the value from n to n+1
    public V getSingleValue(int n) {
        return getValues(new Interval(n, n + 1)).get(0);
    }

    public ArrayList<V> getValues(Interval interval) {
        ArrayList<V> values = new ArrayList<>();
        if (root != null && interval != null)
            getValues(interval, root, values);
        return values;
    }

    private void getValues(Interval interval, TreeNode start, ArrayList<V> results) {
        System.out.println(interval.toString() + ", " + start.toString() + ", " + Arrays.toString(results.toArray()));

        if (interval.fullyLeftOf(start.interval) && start.left != null)
            getValues(interval, start.left, results);
        else if (interval.fullyRightOf(start.interval) && start.right != null)
            getValues(interval, start.right, results);
        else if (interval.isContainedBy(start.interval))
            results.add(start.value); // base case
        else if (interval.contains(start.interval)) {
            results.add(start.value); 
            Interval intLeft = new Interval(interval.low, start.interval.low);
            Interval intRight = new Interval(interval.high, start.interval.high);
            if (start.left != null && intLeft.isValid())
                getValues(intLeft, start.left, results);
            if (start.right != null && intRight.isValid())
                getValues(intRight, start.right, results);
        } else if (interval.partiallyContainsLeft(start.interval)) {
            results.add(start.value);
            Interval newInt = new Interval(start.interval.high, interval.high);
            if (start.right != null && newInt.isValid())
                getValues(newInt, start.right, results);
        } else if (interval.partiallyContainsRight(start.interval)) {
            results.add(start.value);
            Interval newInt = new Interval(interval.low, start.interval.low);
            if (start.left != null && newInt.isValid())
                getValues(newInt, start.left, results);
        }
    }

    public TreeNode constrainIntervals(TreeNode start, Interval i) {
        if (start == null)
            return null;

        start.interval.high = Math.min(start.interval.high, i.high);
        start.interval.low = Math.max(start.interval.low, i.low);

        if (!start.interval.isValid())
            return null;

        if (start.left != null)
            start.setLeft(constrainIntervals(start.left, new Interval(start.left.interval.low, start.interval.low)));
        if (start.right != null)
            start.setRight(
                    constrainIntervals(start.right, new Interval(start.interval.high, start.right.interval.high)));

        return start;
    }

    public void print() {
        System.out.println("================");
        print(root, 0);
        System.out.println("================");
    }

    private void print(TreeNode tree, int level) {
        for (int i = 0; i < level; i++)
            System.out.print("\t");
        if (level != 0)
            System.out.print("|");
        for (int i = 0; i < level * 4; i++)
            System.out.print("-");

        if (tree == null)
            System.out.println("NULL");
        else {
            System.out.println(tree.interval.toString() + ": " + tree.value);
            print(tree.left, level + 1);
            print(tree.right, level + 1);
        }
    }

    boolean isEmpty() {
        return size == 0;
    }

    public class TreeNode {

        V value;
        Interval interval;
        TreeNode left, right;

        public TreeNode(Interval interval_, V value_, TreeNode left_, TreeNode right_) {
            this.setInterval(interval_);
            this.setValue(value_);
            this.setLeft(left_);
            this.setRight(right_);
        }

        public String toString() {
            return this.interval.toString() + ": " + this.value.toString();
        }

        public void setInterval(int low_, int high_) {
            this.interval.set(low_, high_);
        }

        public void setInterval(Interval i) {
            this.interval = i;
        }

        public Interval getInterval() {
            return this.interval;
        }

        public void setValue(V value_) {
            this.value = value_;
        }

        public V getValue() {
            return this.value;
        }

        public void setLeft(TreeNode left_) {
            this.left = left_;
        }

        public void setRight(TreeNode right_) {
            this.right = right_;
        }

        public TreeNode getLeft() {
            return this.left;
        }

        public TreeNode getRight() {
            return this.right;
        }

    }

}
